import Utils from "../Utils";
import PolygonTextureAssembler from "./PolygonTextureAssembler";

const { ccclass, property, mixins, executeInEditMode, menu } = cc._decorator;


let _vec2_temp = new cc.Vec2();
let _mat4_temp = new cc.Mat4();
@ccclass
@executeInEditMode
@mixins(cc.BlendFunc)
@menu('Blade/PolygonTexture')
export default class PolygonTexture extends cc.RenderComponent {


    @property(cc.Texture2D)
    _texture: cc.Texture2D = null;

    @property(cc.Texture2D)
    get texture() {
        return this._texture;
    }
    set texture(val: cc.Texture2D) {
        this._texture = val;
        let l = -val.width / 2, b = -val.height / 2, t = val.height / 2, r = val.width / 2;
        this.polygon = [cc.v2(l, b), cc.v2(r, b), cc.v2(r, t), cc.v2(l, t)];
        this.node.width = val.width;
        this.node.height = val.height;
        this._updateMaterial();
    }
    @property({ type: cc.Enum(cc.macro.BlendFactor), override: true })
    srcBlendFactor: cc.macro.BlendFactor = cc.macro.BlendFactor.SRC_ALPHA;

    @property({ type: cc.Enum(cc.macro.BlendFactor), override: true })
    dstBlendFactor: cc.macro.BlendFactor = cc.macro.BlendFactor.ONE_MINUS_SRC_ALPHA;

    @property({ type: [cc.Vec2], serializable: true })
    _polygon: cc.Vec2[] = [];
    @property({ type: [cc.Vec2], serializable: true })
    public get polygon() {
        return this._polygon;
    }
    public set polygon(points: cc.Vec2[]) {
        this._polygon = points;
        this.setVertsDirty();
    }

    /**
     * @polygonGizmo的开关
     * @需要在packages/polygonpoints的package.json中配置对应target
     */
    @property(cc.Boolean)
    editing: boolean = true;




    onLoad() {
        this.node.on(cc.Node.EventType.TOUCH_MOVE, this.onTouchMove, this);
        this.node['_hitTest'] = this._hitTest.bind(this);
    }

    onDestroy() {
        this.node.off(cc.Node.EventType.TOUCH_MOVE, this.onTouchMove, this);
    }

    /**
     * @hitTest设计到节点的点击事件
     * @param cameraPt 
     * @returns 
     */
    _hitTest(cameraPt: cc.Vec2) {
        let node = this.node;
        let size = node.getContentSize(),
            w = size.width,
            h = size.height,
            testPt = _vec2_temp;

        node['_updateWorldMatrix']();
        // If scale is 0, it can't be hit.
        if (!cc.Mat4.invert(_mat4_temp, node['_worldMatrix'])) {
            return false;
        }
        cc.Vec2.transformMat4(testPt, cameraPt, _mat4_temp);
        return Utils.isInPolygon(testPt, this.polygon);
    }

    onTouchMove(t: cc.Touch) {
        let delta = t.getDelta();
        this.node.setPosition(this.node.getPosition().add(delta));
    }

    /**
     * @在renderComponent的__preload函数中被调用
     * @__preload函数在mainLoop的 node-activator.js @中被调用
     */
    _resetAssembler() {
        /**@自定义assembler的初始化 */
        this._assembler = new PolygonTextureAssembler();
        this._assembler.init(this);
        this._updateColor();
        this.setVertsDirty();
    }

    /**
     * @材质改变时需要调用此函数
     */
    _updateMaterial() {
        let texture = this._texture;
        let material = this.getMaterial(0);
        if (material) {
            if (material.getDefine("USE_TEXTURE") !== undefined) {
                material.define("USE_TEXTURE", true);
            }
            material.setProperty("texture", texture);
        }
        this['_updateBlendFunc']();
        this.setVertsDirty();
    }
}
