import Utils from "../Utils";
import PolygonTexture from "./PolygonTexture";

const { ccclass, property } = cc._decorator;
const gfx = cc['gfx'];

let vfmtPosUvColor = new gfx.VertexFormat([
    /**@节点的世界坐标 @占2个float32 */
    { name: gfx.ATTR_POSITION, type: gfx.ATTR_TYPE_FLOAT32, num: 2 },
    /**@节点的uv坐标 @占2个float32 @如果节点使用了独立的纹理 （未合图） @这里的uv值通常是0或1  @uv坐标原点是左上角 */
    { name: gfx.ATTR_UV0, type: gfx.ATTR_TYPE_FLOAT32, num: 2 },
    /**@节点的颜色值 rgba 占4个uint8 = 1个float32 */
    { name: gfx.ATTR_COLOR, type: gfx.ATTR_TYPE_UINT8, num: 4, normalize: true },
]);
@ccclass
export default class PolygonTextureAssembler extends cc.Assembler {
    constructor() {
        super();
        this._renderData = new cc.RenderData();
        this._renderData.init(this);

        this.initData();
    }
    /**
     * @texture顶点个数 
     * @因为我们是多边形所以这个顶点个数不是固定的4个 
    */
    verticesCount = 4;
    /**
     * @对应三角形的三个点的索引个数 
     * @即 indicesCount = 三角形个数*3 
     * @因为是多边形这个
    */
    indicesCount = 6;
    /**@每个顶点占用float32个数 */
    floatsPerVert = 5;
    /**@uv坐标数组下标相对于数组0位置的偏移量 @共5个位置 0,1存放pos 2,3存放uv 4存放颜色rgba */
    uvOffset = 2;
    /**@顶点颜色数组下标相对于数组0位置的偏移量 @共5个位置 0,1存放pos 2,3存放uv 4存放颜色rgba */
    colorOffset = 4;

    private _renderData: cc.RenderData = null;

    /**@返回当前texture占用的float32个数 */
    get verticesFloats() {
        return this.verticesCount * this.floatsPerVert;
    }

    /**
     * 顶点存储格式
     */
    getVfmt() {
        return vfmtPosUvColor;
    }

    getBuffer() {
        /** 
         * cc.render @提供基础渲染接口的渲染器对象 
         * @所有的Assembler2D数据都会保存到MeshBuffer中 
         * @MeshBuffer中保存着所有的顶点数据包括不同的纹理
        */
        return cc.renderer['_handle'].getBuffer('mesh', this.getVfmt());
    }

    /**
     * @初始化renderData
     */
    initData() {
        let data = this._renderData;
        /**
         * @createQuadData干了2件事
         * @初始化数组存储单位
         * @初始化四个顶点拆分的三角形索引 @因为我们的是多边形所以这个顶点索引会在initQuadIndices中被修改
         */
        data.createQuadData(0, this.verticesFloats, this.indicesCount);
    }

    /**@重置renderData */
    resetData(comp: PolygonTexture) {
        let points = comp.polygon;
        if (!points || points.length < 3) return;
        this.verticesCount = points.length;
        /**@多边形可依次构成的三角形个数 */
        let triangleNum = (this.verticesCount - 2);
        /**@对应三角形顶点个数 */
        this.indicesCount = triangleNum * 3;
        /**@顶点数据改变后需要清空之前旧的renderData创建新的renderData */
        this._renderData['clear']();
        /**@创建新的renderData */
        this.initData();
    }

    /**@初始化三角形顶点索引 */
    initQuadIndices(indices: number[], arr: number[]) {
        for (let i = 0; i < arr.length; i++) {
            indices[i] = arr[i];
        }
    }

    /** 填充顶点的color */
    updateColor(comp: PolygonTexture, color: number) {
        let uintVerts = this._renderData.uintVDatas[0];
        if (!uintVerts) return;
        color = color != null ? color : comp.node.color['_val'];
        let floatsPerVert = this.floatsPerVert;
        let colorOffset = this.colorOffset;

        let polygon = comp.polygon;
        for (let i = 0; i < polygon.length; i++) {
            uintVerts[colorOffset + i * floatsPerVert] = color;
        }
    }
    /** 更新uv */
    protected updateUVs(comp: PolygonTexture) {
        let uvOffset = this.uvOffset;
        let floatsPerVert = this.floatsPerVert;
        let verts = this._renderData.vDatas[0];

        let uvs = Utils.computeUv(comp.polygon, comp.node.width, comp.node.height)
        let polygon = comp.polygon;
        for (let i = 0; i < polygon.length; i++) {
            let dstOffset = floatsPerVert * i + uvOffset;
            verts[dstOffset] = uvs[i].x;
            verts[dstOffset + 1] = uvs[i].y;
        }
    }

    /**@更新世界坐标 */
    protected updateWorldVertsWebGL(comp: PolygonTexture) {
        let verts = this._renderData.vDatas[0];

        let matrix: cc.Mat4 = comp.node['_worldMatrix'];
        /** 
        * @四维矩阵
           [
               a, c, 0, 0,
               b, d, 0, 0,
               0, 0, 1, 0,
               tx, ty, tz, 1
           ]
        */
        let matrixm = matrix.m,
            /**@abcd控制旋转缩放倾斜 */
            a = matrixm[0], b = matrixm[1], c = matrixm[4], d = matrixm[5],
            /**@tx对应世界坐标x偏移量 @ty对应世界坐标的y偏移量 */
            tx = matrixm[12], ty = matrixm[13];

        let justTranslate = a === 1 && b === 0 && c === 0 && d === 1;
        let floatsPerVert = this.floatsPerVert;
        /**
         * @世界坐标平移公式
         * @x = a*px + c*py + tx
         * @y = b*px + d*py + ty
        */
        if (justTranslate) {
            let polygon = comp.polygon;
            for (let i = 0; i < polygon.length; i++) {
                verts[i * floatsPerVert] = polygon[i].x + tx;
                verts[i * floatsPerVert + 1] = polygon[i].y + ty;
            }
        } else {
            let polygon = comp.polygon;
            for (let i = 0; i < polygon.length; i++) {
                verts[i * floatsPerVert] = a * polygon[i].x + c * polygon[i].y + tx;
                verts[i * floatsPerVert + 1] = b * polygon[i].x + d * polygon[i].y + ty;
            }
        }
    }

    protected updateWorldVertsNative(comp: PolygonTexture) {
        let verts = this._renderData.vDatas[0];
        let floatsPerVert = this.floatsPerVert;

        let polygon = comp.polygon;
        for (let i = 0; i < polygon.length; i++) {
            verts[i * floatsPerVert] = polygon[i].x;
            verts[i * floatsPerVert + 1] = polygon[i].y;
        }
    }

    protected updateWorldVerts(comp: PolygonTexture) {
        if (CC_NATIVERENDERER) {
            this.updateWorldVertsNative(comp);
        } else {
            this.updateWorldVertsWebGL(comp);
        }
    }

    /** 更新顶点数据 */
    protected updateVerts(comp: PolygonTexture) {
        let indicesArr = Utils.splitePolygon(comp.polygon);
        this.initQuadIndices(this._renderData.iDatas[0], indicesArr);
        this.updateWorldVerts(comp);
    }

    /** 
     * @更新renderdata 
     * @只有在渲染数据改变时才会调用并不是每帧都调用
     * @详见RenderComponent的setVertsDirty函数
     * */
    protected updateRenderData(comp: PolygonTexture) {
        if (comp._vertsDirty) {
            this.resetData(comp);
            this.updateUVs(comp);
            this.updateVerts(comp);
            this.updateColor(comp, null);
            comp._vertsDirty = false;
        }
    }

    //每帧都会被调用
    fillBuffers(comp: PolygonTexture, renderer) {
        if (renderer.worldMatDirty) {
            this.updateWorldVerts(comp);
        }

        let renderData = this._renderData;

        // vData里包含 pos， uv， color数据， iData中包含顶点索引
        let vData = renderData.vDatas[0];
        let iData = renderData.iDatas[0];

        let buffer = this.getBuffer();
        let offsetInfo = buffer.request(this.verticesCount, this.indicesCount);

        // buffer data may be realloc, need get reference after request.

        // fill vertices
        let vertexOffset = offsetInfo.byteOffset >> 2,
            vbuf = buffer._vData;
        if (vData.length + vertexOffset > vbuf.length) {
            vbuf.set(vData.subarray(0, vbuf.length - vertexOffset), vertexOffset);
        } else {
            vbuf.set(vData, vertexOffset);
        }

        // fill indices
        let ibuf = buffer._iData,
            indiceOffset = offsetInfo.indiceOffset,
            vertexId = offsetInfo.vertexOffset;
        for (let i = 0, l = iData.length; i < l; i++) {
            ibuf[indiceOffset++] = vertexId + iData[i];
        }
    }
}
